"""
As this library is still in early development, the API might change.
I suggest you vendor this so that you don't create conflicts if other
libraries decide to use an older, incompatible version

If you are shipping python source code, then I've included the license
at the bottom of this file to make your life easier.
"""

from packaging.version import Version
from functools import wraps
import inspect

from warnings import warn

from .merge_docstrings import merge_docstrings


def default_parameter_change(version, old_kwargs,
                             library_name, current_library_version):
    """Deprecates a default value for a kwarg.

    If the software version is greater or equal to that of ``version``, this
    decorator returns the original function without any modifications.

    If the user doesn't specify a parameter that will change default values,
    it will issue a ``FutureWarning`` pointing to the line of his code,
    in a single warning for all changes in the default parameters.


    It also modifies the ``__signature__`` of the function so that the
    function appears to have the default parameters to whatever the default
    values are in the current version of the function.

    Finally, if a docstring is provided, it appends a warning
    message compatible numpydoc.

    Parameters
    ----------
    version: version-like
        The version in which the parameter will take the new default value.
        If the software version reaches this value, the new default will be
        taken on without warning.

    old_kwargs:
        The list of keyword arguments, with their old default values
        specified, as a dictionary. You should be able to
        copy paste your old keyword arguments and put then in a `dict`.

    library_name : str
        The human readable name for your library.

    current_library_version : version-like
        The current version of your library.

    Examples
    --------
    >>> @default_parameter_change('0.16', doc(this='tim'))
    ... def foo(this='foo', bar='zip'):
    ...     '''Prints your parameter.
    ...
    ...     **This function has a depeprecation decorator set to '0.16'**
    ...
    ...     Parameters
    ...     ----------
    ...     this : str
    ...         This is the string parameter that should be printed
    ...
    ...     '''
    ...     print(bar)
    ...     return this

    """
    def the_decorator(func):
        if Version(current_library_version) >= Version(version):
            return func

        new_signature = inspect.signature(func)
        base_message = ('In release {version} of {module}, the function '
                        '``{funcname}`` '
                        'will have new default parameters. To avoid this '
                        'warning specify the value of all listed arguments.'
                        '\n\n'.format(version=version, module=library_name,
                                      funcname=func.__name__))

        old_parameters = [
            inspect.Parameter(
                key, param.kind,
                default=param.default
                if key not in old_kwargs else old_kwargs[key])
            for key, param in new_signature.parameters.items()]
        func_args = inspect.getfullargspec(func).args
        old_signature = new_signature.replace(parameters=old_parameters)

        @wraps(func)
        def wrapper(*args, **kwargs):
            issue_warning = False
            message = base_message
            for key, old_value in old_kwargs.items():
                # Check if the argument was provided as a
                # positional argument
                if (new_signature.parameters[key].kind is
                        inspect._POSITIONAL_OR_KEYWORD):
                    arg_pos = func_args.index(key)
                    if len(args) > arg_pos:
                        continue

                # check if it was provided as a kwarg
                if key in kwargs:
                    continue
                new_value = new_signature.parameters[key].default
                message = (
                    f'{message}    The default value of ``{key}`` '
                    f'will be changed from ``{repr(old_value)}`` to '
                    f'``{repr(new_value)}``. '
                    '\n'
                )
                kwargs[key] = old_value
                issue_warning = True
            if issue_warning:
                warn(message, FutureWarning, stacklevel=2)

            return func(*args, **kwargs)

        wrapper.__signature__ = old_signature

        # If the wrapper doesn't have a doc string, don't bother adding one.
        if wrapper.__doc__ is None:
            return wrapper

        doc_deprecated_kwargs = ''
        # Iterating through a list is tedius, though it is the only way to
        # ensure things stay in order in python <=3.6
        for param in old_parameters:
            key = param.name
            if key not in old_kwargs:
                continue
            old_value = param.default
            new_value = new_signature.parameters[key].default
            doc_deprecated_kwargs = (
                f'{doc_deprecated_kwargs}    `{key}` : '
                f'`{repr(old_value)}` -> `{repr(new_value)}`'
                '\n\n'
            )
        warnings_string = f"""
Warns
-----
FutureWarning
  In release {version} of {library_name}, this function will take on
  new default value(s) for the following keyword argument(s):

""" + doc_deprecated_kwargs + """

  To avoid this warning in your code, specify the value of all listed
  keyword arguments.

"""
        wrapper.__doc__ = merge_docstrings(wrapper, warnings_string)
        return wrapper
    return the_decorator


"""
BSD 3-Clause License

Copyright (c) 2018, Mark Harfouche
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this
  list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.

* Neither the name of the copyright holder nor the names of its
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""



"""
As this library is still in early development, the API might change.
I suggest you vendor this so that you don't create conflicts if other
libraries decide to use an older, incompatible version

If you are shipping python source code, then I've included the license
at the bottom of this file to make your life easier.
"""

from packaging.version import Version
from functools import wraps
import inspect
from warnings import warn
from .merge_docstrings import merge_docstrings

POSITIONAL_OR_KEYWORD = inspect.Parameter.POSITIONAL_OR_KEYWORD


# Keep the order of params as is,
# `previous_arg_order` and `keep_old_signature`
# should be optional positional args.
def kwonly_change(version, previous_arg_order=None, keep_old_signature=False,
                  library_name=None, current_library_version=None):
    """Returns a decorator that enforces a smaller number of positional arguments.

    If the current library version is smaller than version in which the new
    signature takes effect, this deprecator will:

    Issue a warnign pointing to the user's code if they call a function with
    too many positional arguments.

    If you want to move the new keyword only arguments after the keyword
    only seperator, you should specify a list with the previous order of
    arguments as the parameter ``previous_arg_order``. If this parameter is
    not specified, it is assumed that the previous function signature allowed
    all parameters to be specified as either positional or keyword arguments.

    Parameters
    ----------
    version: version-like
        Version in which the new function signature takes full effect.

    previous_arg_order: list of strings or ``None``
        If the function previously had keyword only arguments, you should use
        this to specify the names of previous positional arguments. If set
        to none, it is assumed that the previous version of the function
        would accept all parameters as positional or keyword arguments.
        Specifying this also allows you to re-order the new keyword only.
        To re-iterate, you cannot swap the order of positional arguments with
        this deprecator.

    keep_old_signature: bool
        If set to true, the signature will reflect the previous signature of
        the function. This is not recommended since showing new users the
        keyword-only version of the signature will not cause them to write
        wrong code. If they follow the new signature, their code will be
        correct and they will avoid the ``FutureWarning``.

    library_name: str
        Friendly library name to include in warning messages

    current_library_version: version-like
        The current version of the shipped library (typically your module's
        ``__version__``).

    Examples
    --------

    >>> from wabisabi import kwonly_change
    >>> import functools
    >>>
    >>> kwonly_change = functools.partial(kwonly_change,
    ...                                   library_name='my super lib',
    ...                                   current_library_version='0.14')
    >>> @kwonly_change('0.15')
    ... def foo_new(zap='zip', *, bar='hello', baz='world'):
    ...     return zap + bar + baz

    """
    if current_library_version is None:
        raise ValueError('Must provide a version for your library.')

    if library_name is None:
        raise ValueError('You must provide a user friendly name for your '
                         'library.')

    def the_decorator(func):
        if Version(current_library_version) >= Version(version):
            return func
        new_signature = inspect.signature(func)

        new_arg_names = [name for name in new_signature.parameters]

        if previous_arg_order is None:
            old_nargs = len(new_signature.parameters)
            old_arg_names = new_arg_names[:old_nargs]
        else:
            old_nargs = len(previous_arg_order)
            old_arg_names = previous_arg_order
        # These arguments are still required as argument only keywords
        func_args = inspect.getfullargspec(func).args
        new_nargs = len(func_args)

        old_parameters = []
        all_params = {**new_signature.parameters}
        for key in old_arg_names:
            param = all_params.pop(key)
            if key in func_args:
                kind = param.kind
            else:
                kind = POSITIONAL_OR_KEYWORD

            old_parameters.append(
                inspect.Parameter(key, kind=kind,
                                  default=param.default,
                                  annotation=param.annotation))
        for key, param in all_params.items():
            old_parameters.append(
                inspect.Parameter(key, kind=param.kind,
                                  default=param.default,
                                  annotation=param.annotation))

        old_signature = new_signature.replace(parameters=old_parameters)

        @wraps(func)
        def wrapper(*args, **kwargs):

            if len(args) > old_nargs:
                # The warning should be issued here too!
                raise TypeError('{name}() takes {old_nargs} positional '
                                'arguments but {len_args} were given'
                                ''.format(name=func.__name__,
                                          old_nargs=old_nargs,
                                          len_args=len(args)))

            if len(args) > new_nargs:
                for key, value in zip(old_arg_names[new_nargs:len(args)],
                                      args[new_nargs:]):
                    if key in kwargs:
                        calling_function = inspect.stack()[1]

                        s = SyntaxError(
                            "In version {version} of {library_name}, the "
                            "argument ('{key}') has "
                            "will become a keyword only argument. You "
                            "specified it as both a positional argument and "
                            "a keyword argument."
                            "".format(version=version,
                                      library_name=library_name,
                                      key=key))
                        s.lineno = calling_function.lineno
                        s.filename = calling_function.filename
                        # Even the normal syntax errors suck at telling you
                        # the position of your error whe a statement spans
                        # multiple lines
                        # s.offset =  # is it worth it to find where?
                        raise s
                    kwargs[key] = value

                warn("In version {version} of {library_name}, the "
                     "argument(s): '{old_pos_args}' will become keyword-only "
                     "argument(s). To suppress this warning, specify all "
                     "listed argument(s) with keywords."
                     "".format(
                         version=version, library_name=library_name,
                         old_pos_args=old_arg_names[new_nargs:len(args)]),
                     FutureWarning, stacklevel=2)

                args = args[:new_nargs]

            return func(*args, **kwargs)

        if keep_old_signature:
            wrapper.__signature__ = old_signature

        # only add a docstring if they had one already
        if wrapper.__doc__ is None:
            return wrapper

        args = ', '.join(old_arg_names[new_nargs:])
        warnings_string = f"""
Warns
-----
FutureWarning
  In release {version} of {library_name}, the argument(s):

    `{args}`

  will become keyword-only arguments. To avoid this warning,
  provide all the above arguments as keyword arguments.

"""
        wrapper.__doc__ = merge_docstrings(wrapper, warnings_string)
        return wrapper
    return the_decorator


"""
BSD 3-Clause License

Copyright (c) 2018, Mark Harfouche
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this
  list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.

* Neither the name of the copyright holder nor the names of its
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""



from wabisabi import kwarg_name_change
from functools import partial
from pytest import warns, raises
import warnings

__version__ = '0.14'
kwarg_name_change = partial(kwarg_name_change,
                            current_library_version=__version__,
                            library_name='mylib')


@kwarg_name_change('0.15', {'b': 'a'})
def foo_easy(*, a=6):
    return a


def test_too_easy():
    assert foo_easy() == 6
    with warns(FutureWarning, match='In version 0.15'):
        assert foo_easy(b=3) == 3
    with warnings.catch_warnings():
        warnings.simplefilter("error")
        assert foo_easy(a=7) == 7


def test_too_easy_doubly_specified():
    with warns(FutureWarning, match='In version 0.15'):
        with raises(TypeError,
                    match="'b' and 'a' refer to the same parameter."):
            foo_easy(b=1, a=6)

        with raises(TypeError,
                    match="'b' and 'a' refer to the same parameter."):
            foo_easy(a=6, b=1)


@kwarg_name_change('0.15', {'baz': 'b', 'dare': 'd'})
def foo_medium(a, b=2, *, c=6, d=2):
    return a * b * c * d


def test_foo_medium():
    with raises(TypeError, match=r'missing 1 required positional argument'):
        foo_medium()
    assert foo_medium(0) == 0
    with warnings.catch_warnings():
        # renamed kwargs can still be specified as positional arguments
        warnings.simplefilter("error")
        assert foo_medium(1, 1, c=1) == 2

    # Old kwargs give warnings
    with warns(FutureWarning, match='In version 0.15'):
        assert foo_medium(1, baz=1, c=1, dare=1) == 1
    with warns(FutureWarning, match='In version 0.15'):
        assert foo_medium(1, b=1, c=1, dare=1) == 1
    with warns(FutureWarning, match='In version 0.15'):
        assert foo_medium(1, baz=1, c=1, d=1) == 1

    with warnings.catch_warnings():
        # renamed kwargs can still be specified as positional arguments
        warnings.simplefilter("error")
        assert foo_medium(1, b=1, c=1, d=1) == 1

    with warnings.catch_warnings():
        warnings.simplefilter("error")
        assert foo_medium(1, d=2, c=2) == 8

    with warnings.catch_warnings():
        warnings.simplefilter("error")
        assert foo_medium(1, c=2) == 8

    with warnings.catch_warnings():
        warnings.simplefilter("error")
        assert foo_medium(1, 1, c=2) == 4

    with warnings.catch_warnings():
        warnings.simplefilter("error")
        assert foo_medium(1, 1, d=1) == 6


def foo_hard_new(a, b=2, *, c=6, d=2):
    return a * b * c * d


def foo_hard_old(a, b=2, d=2, *, c=6):
    return a * b * c * d


def foo_easy_already_deprecated(*, a=6):
    return a


def test_raises():
    from wabisabi import kwarg_name_change

    # Without a current_library_version
    with raises(ValueError):
        kwarg_name_change('15', library_name='m')(foo_easy_already_deprecated)

    # Without a library_name
    with raises(ValueError):
        kwarg_name_change('15', current_library_version='0.1')(
            foo_easy_already_deprecated)


def test_already_deprecated():
    foo_easy_should_not_change = kwarg_name_change('0.1', {'bar': 'a'})(
        foo_easy_already_deprecated)
    assert foo_easy_should_not_change == foo_easy_already_deprecated


@kwarg_name_change('0.15', {'bar': 'b'})
def foo_docs(a='a', *, b='b'):
    """This is foo

    Parameters
    ----------
    a: str
        This is a.

    b: str
        This is b

    """
    return a + b


def test_doc():
    assert "Warns" in foo_docs.__doc__
    assert "'b' refers to the old keyword argument 'bar'" in foo_docs.__doc__



"""
As this library is still in early development, the API might change.
I suggest you vendor this so that you don't create conflicts if other
libraries decide to use an older, incompatible version

If you are shipping python source code, then I've included the license
at the bottom of this file to make your life easier.
"""

from packaging.version import Version
from functools import wraps
from warnings import warn
from .merge_docstrings import merge_docstrings


# Keep the order of params as is,
# `previous_arg_order` and `keep_old_signature`
# should be optional positional args.
def kwarg_name_change(version, previous_kwarg_map=None,
                      library_name=None, current_library_version=None):
    """A decorator factory for facilitating keyword arguments name changes.

    If the current library version is smaller than version in which the new
    signature takes effect, this deprecator will:

    Issue a warnign pointing to the user's code if they call a function with
    too many positional arguments.

    Parameters
    ----------
    version: version-like
        Version in which the new function signature takes full effect.

    previous_kwarg_map: dictionary of kwarg name maps
        This dictionary contains the keys (the old kwarg names) and the
        associated values of the new kwarg names. You may specifiy as
        many kwarg name pairs as you wish.

    library_name: str
        Friendly library name to include in warning messages

    current_library_version: version-like
        The current version of the shipped library (typically your module's
        ``__version__``).

    Examples
    --------

    >>> from wabisabi import kwonly_change
    >>> import functools
    >>>
    >>> kwarg_name_change = functools.partial(kwarg_name_change,
    ...                                       library_name='my super lib',
    ...                                       current_library_version='0.14')
    >>> @kwarg_name_change('0.15', {'bar': 'foo'})
    ... def foo_new(zap='zip', *, foo='hello', baz='world'):
    ...     return zap + foo + baz

    """
    if current_library_version is None:
        raise ValueError('Must provide a version for your library.')

    if library_name is None:
        raise ValueError('You must provide a user friendly name for your '
                         'library.')

    def the_decorator(func):
        if Version(current_library_version) >= Version(version):
            return func

        @wraps(func)
        def wrapper(*args, **kwargs):
            new_kwargs = {}
            for key, value in kwargs.items():
                if key not in previous_kwarg_map:
                    new_key = key
                else:
                    new_key = previous_kwarg_map[key]

                    warn(f"In version {version} of {library_name}, the "
                         f"keyword argument '{key}' will be replaced by "
                         f"'{new_key}'. "
                         "To suppress this warning, specify the keyword "
                         "argument with its new name.",
                         FutureWarning, stacklevel=2)
                if new_key in new_kwargs:
                    # Search for the old key.
                    for old_key, n_key in previous_kwarg_map.items():
                        if n_key == new_key:
                            break

                    raise TypeError(
                        "'{old_key}' and '{new_key}' refer to the same "
                        "parameter. Using both is not allowed."
                        "".format(old_key=old_key, new_key=new_key))
                new_kwargs[new_key] = value
            return func(*args, **new_kwargs)

        # only add a docstring if they had one already
        if wrapper.__doc__ is None:
            return wrapper

        warnings_string = """
Warns
-----
FutureWarning
    In release {version} of {module}, the following keyword arguments can only
    be referred by their new names. For now, both names are accepted as valid:

""".format(version=version, module=library_name)

        for key, new_key in previous_kwarg_map.items():
            warnings_string += (f"""
        '{new_key}' refers to the old keyword argument '{key}'""")

        warnings_string += """

    To avoid this warning, provide all the listed keyword arguments using their
    new names.

"""

        wrapper.__doc__ = merge_docstrings(wrapper, warnings_string)
        return wrapper
    return the_decorator


"""
BSD 3-Clause License

Copyright (c) 2019, Mark Harfouche
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this
  list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.

* Neither the name of the copyright holder nor the names of its
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""



import numpydoc
from packaging.version import Version
from numpydoc.docscrape import FunctionDoc as _FunctionDoc
from numpydoc.docscrape import NumpyDocString  # noqa

try:
    if Version(numpydoc.__version__) < Version('0.7'):
        raise ImportError('numpydoc mangling requires numpydoc 0.7.')
except AttributeError:
    raise ImportError('numpydoc mangling requires numpydoc 0.7.')


class FunctionDoc(_FunctionDoc):

    def _str_index(self):
        # This should not be necessary after
        # See https://github.com/numpy/numpydoc/pull/187/files
        idx = self['index']
        out = []
        output_index = False
        default_index = idx.get('default', '')
        if default_index:
            output_index = True
        out += ['.. index:: %s' % default_index]
        for section, references in idx.items():
            if section == 'default':
                continue
            output_index = True
            out += ['   :%s: %s' % (section, ', '.join(references))]
        if output_index:
            return out
        else:
            return ''

    def _str_see_also(self, func_role):
        # See https://github.com/numpy/numpydoc/pull/189
        if not self['See Also']:
            return []
        out = []
        out += self._str_header("See Also")
        last_had_desc = True
        for func, desc, role in self['See Also']:
            if role:
                link = ':%s:`%s`' % (role, func)
            elif func_role:
                link = ':%s:`%s`' % (func_role, func)
            else:
                link = "%s" % func
            if desc or last_had_desc:
                out += ['']
                out += [link]
            else:
                out[-1] += ", %s" % link
            if desc:
                out += self._str_indent([' '.join(desc)])
                last_had_desc = True
            else:
                last_had_desc = False
        out += ['']
        return out

    def __str__(self, func_role=''):
        out = []
        # See https://github.com/numpy/numpydoc/issues/190
        # out += self._str_signature()
        out += self._str_summary()
        out += self._str_extended_summary()
        for param_list in ('Parameters', 'Returns', 'Yields',
                           'Other Parameters', 'Raises', 'Warns'):
            out += self._str_param_list(param_list)
        out += self._str_section('Warnings')
        out += self._str_see_also(func_role)
        for s in ('Notes', 'References', 'Examples'):
            out += self._str_section(s)
        for param_list in ('Attributes', 'Methods'):
            out += self._str_param_list(param_list)
        out += self._str_index()
        return '\n'.join(out)

    def __add__(self, other):
        for section in self.sections:
            if section in ['Signature', 'Summary']:
                # Summary should be
                continue
            elif section == 'index':
                idx = {**other['index']}
                idx.update(self['index'])
                self['index'] = idx
            elif section == 'Extended Summary':
                self['Extended Summary'] += other['Summary']
                self['Extended Summary'] += other['Extended Summary']
            else:
                self[section] += other[section]
        return self


"""

The __str__, _str_index, and __str_see_also methods are slightly modified
versions of the methods that belong to the numpydoc developmenet team.

This copyright applies to those 3 methods listed above.


Copyright (C) 2008 Stefan van der Walt <stefan@mentat.za.net>,
Pauli Virtanen <pav@iki.fi>

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

"""



from wabisabi import kwonly_change
from functools import partial
from pytest import warns, raises
import inspect
import warnings

__version__ = '0.14'
kwonly_change = partial(kwonly_change,
                        current_library_version=__version__,
                        library_name='mylib')


@kwonly_change('0.15')
def foo_easy(*, a=6):
    return a


def test_too_easy():
    assert foo_easy() == 6
    with warns(FutureWarning, match='In version 0.15'):
        assert foo_easy(3) == 3
    with warnings.catch_warnings():
        warnings.simplefilter("error")
        assert foo_easy(a=7) == 7


@kwonly_change('0.15')
def foo_medium(a, b=2, *, c=6, d=2):
    return a * b * c * d


def test_foo_medium():
    with raises(TypeError, match=r'.* missing 1 required positional argument'):
        foo_medium()
    assert foo_medium(0) == 0
    with warns(FutureWarning, match='In version 0.15'):
        assert foo_medium(1, 1, 1) == 2

    with warns(FutureWarning, match='In version 0.15'):
        assert foo_medium(1, 1, 1, 1) == 1

    with warns(FutureWarning, match='In version 0.15'):
        assert foo_medium(1, 1, 1, d=1) == 1

    with warnings.catch_warnings():
        warnings.simplefilter("error")
        assert foo_medium(1, d=2, c=2) == 8

    with warnings.catch_warnings():
        warnings.simplefilter("error")
        assert foo_medium(1, c=2) == 8

    with warnings.catch_warnings():
        warnings.simplefilter("error")
        assert foo_medium(1, 1, c=2) == 4

    with warnings.catch_warnings():
        warnings.simplefilter("error")
        assert foo_medium(1, 1, d=1) == 6


def foo_hard_new(a, b=2, *, c=6, d=2):
    return a * b * c * d


def foo_hard_old(a, b=2, d=2, *, c=6):
    return a * b * c * d


foo_hard_no_sig_change = kwonly_change('0.15',
                                       previous_arg_order=['a', 'b', 'd'])(
                                           foo_hard_new)

foo_hard_sig_change = kwonly_change('0.15',
                                    previous_arg_order=['a', 'b', 'd'],
                                    keep_old_signature=True)(foo_hard_new)


def test_foo_hard():
    with raises(TypeError, match=r'.* missing 1 required positional argument'):
        foo_hard_no_sig_change()

    # Specifying d both as position and keyword only
    with raises(SyntaxError, match='In version 0.15'):
        assert foo_hard_no_sig_change(1, 1, 1, d=1) == 1

    with raises(TypeError, match=r'.* takes 3 positional arguments'):
        foo_hard_no_sig_change(1, 1, 1, 1)

    assert foo_hard_new(a=1, b=2, c=3, d=4) == \
        foo_hard_no_sig_change(a=1, b=2, c=3, d=4)


def test_foo_signature_change():
    assert inspect.signature(foo_hard_sig_change) == \
        inspect.signature(foo_hard_old)
    assert foo_hard_new(a=1, b=2, c=3, d=4) == \
        foo_hard_sig_change(a=1, b=2, c=3, d=4)


def foo_easy_already_deprecated(*, a=6):
    return a


def test_raises():
    from wabisabi import kwonly_change

    # Without a current_library_version
    with raises(ValueError):
        kwonly_change('15', library_name='mylib')(foo_easy_already_deprecated)

    # Without a library_name
    with raises(ValueError):
        kwonly_change('15', current_library_version='0.1')(
            foo_easy_already_deprecated)


def test_already_deprecated():
    foo_easy_should_not_change = kwonly_change('0.1')(
        foo_easy_already_deprecated)
    assert foo_easy_should_not_change == foo_easy_already_deprecated


@kwonly_change('0.15')
def foo_docs(a='a', *, b='b'):
    """This is foo

    Parameters
    ----------
    a: str
        This is a.

    b: str
        This is b

    """
    return a + b


def test_doc():
    assert "Warns" in foo_docs.__doc__
    assert "`b`" in foo_docs.__doc__



__author__ = """Mark Harfouche"""
__email__ = 'mark.harfouche@gmail.com'
from .default_parameter_change import default_parameter_change  # noqa
from .kwonly_change import kwonly_change                        # noqa
from .kwarg_name_change import kwarg_name_change                # noqa
from ._version import __version__                               # noqa



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages


# Loads _version.py module without importing the whole package.
def get_version_and_cmdclass(pkg_path):
    import os
    from importlib.util import module_from_spec, spec_from_file_location
    spec = spec_from_file_location(
        'version', os.path.join(pkg_path, '_version.py'),
    )
    module = module_from_spec(spec)
    spec.loader.exec_module(module)
    return module.__version__, module.get_cmdclass(pkg_path)


version, cmdclass = get_version_and_cmdclass('wabisabi')


with open('README.md') as readme_file:
    readme = readme_file.read()

with open('HISTORY.md') as history_file:
    history = history_file.read()


test_requirements = ['pytest', 'numpydoc']

setup(
    author="Mark Harfouche",
    author_email='mark.harfouche@gmail.com',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ],
    description=(
        "Deprecate your code effortlessly with decorators. "
        "Give useful warnings and docstrings for different deprecations."
    ),
    license="BSD license",
    long_description=readme + '\n\n' + history,
    long_description_content_type='text/markdown',
    include_package_data=True,
    keywords='deprecation factory wabisabi',
    name='wabisabi',
    packages=find_packages(include=['wabisabi']),
    install_requires=['packaging'],
    python_requires=">=3.8",
    setup_requires=['setuptools'],
    tests_require=test_requirements,
    url='https://github.com/hmaarrfk/wabisabi',
    version=version,
    cmdclass=cmdclass,
    zip_safe=False,
)



import re
import textwrap


def merge_docstrings(func, additional_doc):
    try:
        from .docscrape import FunctionDoc, NumpyDocString

        func_docs = FunctionDoc(func)
        additional_doc = NumpyDocString(additional_doc)
        final_docs = func_docs + additional_doc

        return str(final_docs)
    except ImportError:
        return func.__doc__ + guess_indentation(func.__doc__,
                                                additional_doc)


def guess_indentation(function_doc, additional_doc):
    # Not use anymomre, but this was the little logic I used to guess the
    # indentation amount. I think using numpydoc's str function is better.
    parameters_line = re.search('.*Parameters$', function_doc,
                                flags=re.MULTILINE)
    if parameters_line is not None:
        indentation_amount = parameters_line.group().find('P')
        additional_doc = textwrap.indent(
            additional_doc, ' ' * indentation_amount)
    return additional_doc



import pytest
from wabisabi import default_parameter_change
from functools import partial
from pytest import warns


pytest.importorskip('numpydoc')

from numpydoc.docscrape import FunctionDoc  # noqa

__version__ = '0.14'
default_parameter_change = partial(default_parameter_change,
                                   current_library_version=__version__,
                                   library_name='mylib')


def foo_with_docs(bar='hello'):
    """This is a function foo!

    It has some docs, isn't this cool!

    Parameters
    ----------
    bar: str
        This is a parameter.

    """
    return bar


foo_with_docs_15 = default_parameter_change('0.15',
                                            dict(bar='world'))(foo_with_docs)
foo_with_docs_13 = default_parameter_change('0.13',
                                            dict(bar='world'))(foo_with_docs)


def test_foo_docs():
    assert 'Warns' in foo_with_docs_15.__doc__
    docs = FunctionDoc(foo_with_docs_15)
    assert 'Warns' in docs
    assert 'FutureWarning' == docs['Warns'][0].type
    assert "`bar` : `'world'` -> `'hello'`" in '\n'.join(docs['Warns'][0][2])

    assert foo_with_docs_13.__doc__ == foo_with_docs.__doc__

    with warns(FutureWarning, match='In release 0.15 of mylib'):
        assert foo_with_docs_15() == 'world'


def foo2_with_docs(bar='hello', baz='world'):
    """This is a function foo!

    It has some docs, isn't this cool!

    Parameters
    ----------
    bar: str
        This is a parameter.

    """
    return bar + ' ' + baz


foo2_with_docs_15 = default_parameter_change(
    '0.15', dict(bar='bonjour', baz='monde'))(foo2_with_docs)
foo2_with_docs_13 = default_parameter_change(
    '0.13', dict(bar='bonjour', baz='monde'))(foo2_with_docs)


def test_foo2_docs():
    assert 'Warns' in foo2_with_docs_15.__doc__
    docs = FunctionDoc(foo2_with_docs_15)
    assert 'Warns' in docs
    assert 'FutureWarning' == docs['Warns'][0].type
    assert "`bar` : `'bonjour'` -> `'hello'`" in '\n'.join(docs['Warns'][0][2])
    assert "`baz` : `'monde'` -> `'world'`" in '\n'.join(docs['Warns'][0][2])
    assert "`'hello'`\n\n  `baz`" in '\n'.join(docs['Warns'][0][2])
    assert foo2_with_docs_13.__doc__ == foo2_with_docs.__doc__

    with warns(FutureWarning, match='In release 0.15 of mylib'):
        assert foo2_with_docs_15() == 'bonjour monde'
    assert foo2_with_docs_13() == 'hello world'



from wabisabi import default_parameter_change
from functools import partial
import warnings
import inspect
import pytest
from pytest import warns


parametrize = pytest.mark.parametrize
__version__ = '0.14'
versions_above = ['0.14.1', '0.14.post0', '0.15', '1.0']
default_parameter_change = partial(default_parameter_change,
                                   current_library_version=__version__,
                                   library_name='mylib')


def foo(bar='hello'):
    return bar


def test_dummy():
    assert foo() == 'hello'


@parametrize('version_above', versions_above)
def test_deprecating(version_above):
    foo_0_15 = default_parameter_change(version_above, dict(bar='bar'))(foo)
    with warns(FutureWarning,
               match='In release {} of mylib'.format(version_above)) as record:
        assert foo_0_15() == 'bar'

    assert len(record) == 1
    the_warning = record[0]
    assert 'The default value of ``bar``' in str(the_warning.message)
    assert "from ``'bar'`` to ``'hello'``" in str(the_warning.message)


@parametrize('version_above', versions_above)
def test_deprecating_param_provided(version_above):
    foo_0_15 = default_parameter_change(version_above, dict(bar='bar'))(foo)
    with warnings.catch_warnings():
        warnings.simplefilter("error")
        assert foo_0_15(bar='bar23') == 'bar23'

    with warnings.catch_warnings():
        warnings.simplefilter("error")
        assert foo_0_15('bar23') == 'bar23'


def test_deprecated():
    foo_0_13 = default_parameter_change('0.13', dict(bar='bar'))(foo)

    assert inspect.signature(foo) == inspect.signature(foo_0_13)
    assert foo == foo_0_13
    with warnings.catch_warnings():
        warnings.simplefilter("error")
        assert foo_0_13() == 'hello'


def foo2(a, b='b', c='c'):
    return a + b + c


foo2_0_15 = default_parameter_change('0.15', dict(b='B', c='C'))(foo2)
foo2_0_13 = default_parameter_change('0.13', dict(b='B', c='C'))(foo2)


def test_deprecating2():
    with warns(FutureWarning, match='In release 0.15 of mylib') as record:
        # Be careful adding code between these two
        expeccted_warning_lineno = inspect.currentframe().f_lineno + 1
        assert foo2_0_15('a') == 'aBC'

    assert len(record) == 1
    the_warning = record[0]
    assert expeccted_warning_lineno == the_warning.lineno
    assert 'The default value of ``b``' in str(the_warning.message)
    assert "from ``'B'`` to ``'b'``" in str(the_warning.message)

    assert 'The default value of ``c``' in str(the_warning.message)
    assert "from ``'C'`` to ``'c'``" in str(the_warning.message)

    with warns(FutureWarning, match='In release 0.15 of mylib') as record:
        assert foo2_0_15('a', b='D') == 'aDC'

    assert len(record) == 1
    the_warning = record[0]

    assert 'The default value of ``b``' not in str(the_warning.message)
    assert "from ``'B'`` to ``'c'``" not in str(the_warning.message)

    assert 'The default value of ``c``' in str(the_warning.message)
    assert "from ``'C'`` to ``'c'``" in str(the_warning.message)

    # Positional argument
    with warns(FutureWarning, match='In release 0.15 of mylib') as record:
        assert foo2_0_15('a', 'D') == 'aDC'

    assert len(record) == 1
    the_warning = record[0]

    assert 'The default value of ``b``' not in str(the_warning.message)
    assert "from ``'B'`` to ``'c'``" not in str(the_warning.message)

    assert 'The default value of ``c``' in str(the_warning.message)
    assert "from ``'C'`` to ``'c'``" in str(the_warning.message)

    with warns(FutureWarning, match='In release 0.15 of mylib') as record:
        assert foo2_0_15('a', c='D') == 'aBD'

    assert len(record) == 1
    the_warning = record[0]

    assert 'The default value of ``b``' in str(the_warning.message)
    assert "from ``'B'`` to ``'b'``" in str(the_warning.message)

    assert 'The default value of ``c``' not in str(the_warning.message)
    assert "from ``'B'`` to ``'c'``" not in str(the_warning.message)


def test_deprecating_param_provided2():
    with warnings.catch_warnings():
        warnings.simplefilter("error")
        assert foo2_0_15('a', b='X', c='Y') == 'aXY'

    with warnings.catch_warnings():
        warnings.simplefilter("error")
        assert foo2_0_15('a', 'X', 'Y') == 'aXY'


def test_deprecated2():
    assert inspect.signature(foo2) == inspect.signature(foo2_0_13)
    assert foo2 == foo2_0_13
    with warnings.catch_warnings():
        warnings.simplefilter("error")
        assert foo2_0_13('a') == 'abc'


def test_foo2_signature():
    sig_15 = inspect.signature(foo2_0_15)
    assert sig_15.parameters['b'].default == 'B'
    assert sig_15.parameters['c'].default == 'C'

    sig_13 = inspect.signature(foo2_0_13)
    assert sig_13.parameters['b'].default == 'b'
    assert sig_13.parameters['c'].default == 'c'


def foo_with_docs(bar='hello', zip='zap'):
    """This is a function foo!

    It has some docs, isn't this cool!

    Parameters
    ----------
    bar: str
        This is a parameter.

    """
    return bar


# Something needs to use the decorator syntax or I might make
# a tool that can't be used the way I want it to.
@default_parameter_change('0.15', dict(bar='world'))
def foo_with_docs_15(bar='hello', zip='zap'):
    """This is a function foo!

    It has some docs, isn't this cool!

    Parameters
    ----------
    bar: str
        This is a parameter.

    """
    return bar


foo_with_docs_13 = default_parameter_change('0.13', dict(bar='world'))(
    foo_with_docs)


def test_foo_docs():
    assert 'Warns' in foo_with_docs_15.__doc__
    assert 'FutureWarning' in foo_with_docs_15.__doc__
    assert "`bar` : `'world'` -> `'hello'`" in foo_with_docs_15.__doc__
    assert "zip" not in foo_with_docs_15.__doc__

    assert foo_with_docs_13.__doc__ == foo_with_docs.__doc__

    with warns(FutureWarning, match='In release 0.15 of mylib'):
        assert foo_with_docs_15() == 'world'
    assert foo_with_docs() == 'hello'


def foo2_with_docs(bar='hello', baz='world'):
    """This is a function foo!

    It has some docs, isn't this cool!

    Parameters
    ----------
    bar: str
        This is a parameter.

    """
    return bar + ' ' + baz


foo2_with_docs_15 = default_parameter_change(
    '0.15', dict(bar='bonjour', baz='monde'))(foo2_with_docs)
foo2_with_docs_13 = default_parameter_change(
    '0.13', dict(bar='bonjour', baz='monde'))(foo2_with_docs)


def test_foo2_docs():
    assert 'Warns' in foo2_with_docs_15.__doc__
    assert 'FutureWarning' in foo2_with_docs_15.__doc__
    assert "`bar` : `'bonjour'` -> `'hello'`" in foo2_with_docs_15.__doc__
    assert "`baz` : `'monde'` -> `'world'`" in foo2_with_docs_15.__doc__
    assert foo2_with_docs_13.__doc__ == foo2_with_docs.__doc__

    with warns(FutureWarning, match='In release 0.15 of mylib'):
        assert foo2_with_docs_15() == 'bonjour monde'
    assert foo2_with_docs_13() == 'hello world'
